#!/bin/python

license = """
/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 */
 """

import sys
from xml.etree.ElementTree import ElementTree
from optparse import OptionParser
from string import Template

from CGenerator import CGenerator
from PyGenerator import PyGenerator
from LuaGenerator import LuaGenerator

length_enum = dict()
    
class Parameter:
    def __init__(self, parameter):
        self.name = parameter.attrib["name"]
        self.type_size = int(parameter.attrib["type_size"])
        self.type_count = int(parameter.attrib["type_count"])
        self.offset = 0
        self.description = str(parameter.text) if parameter.text else ""
        
        if self.type_count > 1:
            length_enum[self.name.upper() + "_LEN"] = self.type_count
        
    def setOffset(self, offset):
        self.offset = offset
        
    def getLength(self):
        return (self.type_size * self.type_count) / 8
        
    
class Error:
    def __init__(self, error):
        self.id = error.attrib["id"]
        self.value = error.attrib["value"]
        self.documentation = str(error.text)
        
    def getError(self):
        return self.id
    
    def getValue(self):
        return self.value
    
    def getEnumEntry(self):
        return str(self.id) + " = " + str(self.value) + ", /**< " + self.documentation +" */"
        

class ApiCall:
    def __init__(self, name, message_id, request_params, response_params, timeout, modifier_id=0, description=None):
        self.name = name
        self.message_id = message_id
        self.request_params = request_params
        self.response_params = response_params
        self.modifier_id = modifier_id
        self.description = description
        self.timeout = timeout
        
        offset = 0
        if self.request_params:
            for i in self.request_params:
                i.setOffset(offset)
                offset += i.getLength()
            
        offset = 0
        if self.response_params:
            for i in self.response_params:
                i.setOffset(offset)
                offset += i.getLength()
        
class ModifierId:
    def __init__(self, name, value, prefix):
        self.name = name
        self.value = value
        self.prefix = prefix
    
    def get_name(self):
        return self.name
        
    def get_value(self):
        return self.value
        
    def get_id(self):
        return self.prefix + self.name.upper()
        
class MessageId:
    def __init__(self, name, value):
        self.name = name
        
        first = True
        if (name[0:3] == "dSM"):
            name = "Dsm" + name[3:]
        for c in name:
            if (first):
                self.id = c
                first = False
            else:
                if (c.isupper()):
                    self.id += "_" + c
                else:
                    self.id += c.upper()
        self.value = int(value, 16)
        self.prefix = self.id.upper() + "_"
        self.modifiers = list()
        self_timeout = 0
        
    def set_timeout(self, timeout):
        self.timeout = timeout
        
    def get_timeout(self):
        return self.timeout
        
    def get_enum_decl(self):
        return self.id + " = 0x%02x" % self.value
    
    def get_id(self):
        return self.id

    def add_modifier(self, modifier):
        self.modifiers.append(ModifierId(modifier.attrib["identifier"], int(modifier.attrib["value"]), self.prefix))
        return self.modifiers[len(self.modifiers) - 1]
        
    def get_modifier_enum(self):
        if len(self.modifiers) == 0:
            return ""
            
        out = "typedef enum {\n"
        out += "".join(map(lambda x: "    " + x.get_id() + " = " + str(x.get_value()) + ",\n", self.modifiers))
        out += "} " + self.name + "Modifiers_t;"
        return out
        
    def get_prefix(self):
        return self.prefix
        
    def get_name(self):
        return self.name
                
header_clause = license + """/*
 * DO NOT MANUALLY MODIFY! 
 *
 * THIS FILE IS AUTOGENERATED FROM THE dSM-API XML DEFINITIONS IN
 * dsm-api.xml.
 * Make any changes in the XML definition and regenerate this file.
 */
"""

if __name__ == "__main__":
    parser = OptionParser()
    parser.add_option("-i", "--input", dest="infile",
                      help="XML input")
    parser.add_option("-s", "--schema", dest="schema",
                      help="XML schema definition")
    parser.add_option("-d", "--definition", dest="header",
                      help="C header output file")
    parser.add_option("-c", "--code", dest="code",
                      help="C source code output file")
    parser.add_option("-e", "--enums", dest="enums",
                      help="C enum header output file")
    parser.add_option("-p", "--swig-python", dest="swig_python",
                      help="SWIG output file for Python bindings")
    parser.add_option("-l", "--swig-lua", dest="swig_lua",
                      help="SWIG output file for Lua bindings")
    
    (options, args) = parser.parse_args()
    
    try:
        from lxml import etree
        xmlschema_doc = etree.parse(options.schema)
        xmlschema = etree.XMLSchema(xmlschema_doc)
        
        etree_doc = etree.parse(options.infile)
        
        xmlschema.assertValid(etree_doc)
    except etree.DocumentInvalid as e:
        print e
        sys.stderr.write("The XML file: " + options.infile + "is not valid according to the schema definition\n")
        exit(-1)
    except:
        print "lxml not installed. The XML file", options.infile, "is not validated against the schema."
        pass
    
    tree = ElementTree()
    doc = tree.parse(options.infile)
    
    calls = list()
    events = list()
    message_ids = list()
    prefixes = dict()
    error_list = dict()
    
    errors = map(Error, doc.findall("error"))
    for error in errors:
        error_list[int(error.getValue(), 16)] = error
    
    for api_call in doc.findall("dsm-call"):
        message_id = MessageId(api_call.attrib["name"], api_call.attrib["message_id"])
        message_ids.append(message_id)
        
        if message_id.get_prefix() in prefixes:
            sys.stderr.write(message_id.get_name() + " leads to a duplicate prefix with " + prefixes[message_id.get_prefix()] + "\n")
            exit()
        else:
            prefixes[message_id.get_prefix()]= message_id.get_name()
        
        timeouts = api_call.findall("timeout")
        if (timeouts and len(timeouts) > 0):
            message_id.set_timeout(int(timeouts[0].attrib["seconds"]))
        else:
            message_id.set_timeout(0)
        
        modifiers = api_call.findall("modifier")
        common_request = api_call.findall("request/parameter")
        common_response = api_call.findall("response/parameter")
        common_description = api_call.findall("description")
        common_error = api_call.findall("error")
        common_description = str(common_description[0].text) if common_description[0].text else ""
        
        if (len(modifiers) == 0):
            calls.append(ApiCall(message_id.get_name(), 
                                 message_id.get_id(), 
                                 map(Parameter, common_request), 
                                 map(Parameter, common_response), 
                                 message_id.get_timeout(),
                                 0, 
                                 common_description))
        else:
            for modifier in modifiers:
                modifier_obj = message_id.add_modifier(modifier)
                description = modifier.findall("description")
                documentation = common_description
                
                mod_timeouts = modifier.findall("timeout")
                if (mod_timeouts and len(mod_timeouts) > 0):
                    mod_timeout = int(mod_timeouts[0].attrib["seconds"])
                else:
                    mod_timeout = message_id.get_timeout();
                
                documentation += " " + str(description[0].text) if (len(description) > 0 and description[0].text) else ""
                
                calls.append(ApiCall(message_id.get_name() + "_" + modifier.attrib["identifier"], 
                                     message_id.get_id(),
                                     map(Parameter, common_request + modifier.findall("request/parameter")),
                                     map(Parameter, common_response + modifier.findall("response/parameter")),
                                     mod_timeout,
                                     modifier_obj.get_id(),
                                     documentation))
    
    
    for api_event in doc.findall("dsm-event"):
        message_id = MessageId(api_event.attrib["name"], api_event.attrib["message_id"])
        message_ids.append(message_id)
        
        if message_id.get_prefix() in prefixes:
            sys.stderr.write(message_id.get_name() + " leads to a duplicate prefix with " + prefixes[message_id.get_prefix()] + "\n")
            exit()
        else:
            prefixes[message_id.get_prefix()]= message_id.get_name()
            
        common_response = api_event.findall("response/parameter")
        common_description = api_event.findall("description")
        modifiers = api_event.findall("modifier")
        if (len(common_description) > 0):
            common_description = str(common_description[0].text)
        
        if (len(modifiers) == 0):
            events.append(ApiCall(message_id.get_name(), 
                                  message_id.get_id(), 
                                  None, 
                                  map(Parameter, common_response), 
                                  0,
                                  0, 
                                  common_description))
        else:
            for modifier in modifiers:
                modifier_obj = message_id.add_modifier(modifier)
                description = modifier.findall("description")
                documentation = common_description
                
                if (len(description) > 0):
                    documentation += "\n *\n * " + str(description[0].text)
                
                events.append(ApiCall(message_id.get_name() + "_" + modifier.attrib["identifier"], 
                                      message_id.get_id(),
                                      None,
                                      map(Parameter, common_response + modifier.findall("response/parameter")),
                                      0,
                                      modifier_obj.get_id(),
                                      documentation))

    c_generator = CGenerator(header_clause, calls, events)
    c_generator.header(options.header)
    c_generator.implementation(options.code)
    c_generator.const_header(options.enums, message_ids, error_list, length_enum)
    
    if options.swig_python:
        py_generator = PyGenerator(header_clause, calls, events)
        py_generator.swig(options.swig_python, length_enum)
    
    if options.swig_lua:
        lua_generator = LuaGenerator(header_clause, calls, events)
        lua_generator.swig(options.swig_lua, length_enum)
